# katcp.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 SKA South Africa (http://ska.ac.za/)
# BSD license - see COPYING for details

"""Utilities for dealing with KAT device control
language messages.
"""

import threading
import sys
import re
import time
import warnings

SEC_TO_MS_FAC = 1000
MS_TO_SEC_FAC = 1./1000
DEFAULT_KATCP_MAJOR = 5
# First major version to use seconds (in stead of milliseconds) for timestamps
SEC_TS_KATCP_MAJOR = 5
# First major version to support message IDs
MID_KATCP_MAJOR = 5
# First major version to support #version-connect informs
VERSION_CONNECT_KATCP_MAJOR = 5
# First major version to support #interface-changed informs
INTERFACE_CHANGED_KATCP_MAJOR = 5

class KatcpSyntaxError(ValueError):
    """Exception raised by parsers on encountering syntax errors."""

class KatcpClientError(Exception):
    """Raised by KATCP clients when errors occur."""

class KatcpVersionError(KatcpClientError):
    """
    Exception raised when a katcp feature not supported by the katcp version
    of the server/client is requested
    """

class Message(object):
    """Represents a KAT device control language message.

    Parameters
    ----------
    mtype : Message type constant
        The message type (request, reply or inform).
    name : str
        The message name.
    arguments : list of strings
        The message arguments.
    mid : str, digits only
        The message identifier. Replies and informs that
        are part of the reply to a request should have the
        same id as the request did.
    """

    # Message types
    REQUEST, REPLY, INFORM = range(3)

    # Reply codes
    # TODO: make use of reply codes in device client and server
    OK, FAIL, INVALID = "ok", "fail", "invalid"

    ## @brief Mapping from message type to string name for the type.
    TYPE_NAMES = {
        REQUEST: "REQUEST",
        REPLY: "REPLY",
        INFORM: "INFORM",
    }

    ## @brief Mapping from message type to type code character.
    TYPE_SYMBOLS = {
        REQUEST: "?",
        REPLY: "!",
        INFORM: "#",
    }

    # pylint fails to realise TYPE_SYMBOLS is defined
    # pylint: disable-msg = E0602

    ## @brief Mapping from type code character to message type.
    TYPE_SYMBOL_LOOKUP = dict((v, k) for k, v in TYPE_SYMBOLS.items())

    # pylint: enable-msg = E0602

    ## @brief Mapping from escape character to corresponding unescaped string.
    ESCAPE_LOOKUP = {
        "\\": "\\",
        "_": " ",
        "0": "\0",
        "n": "\n",
        "r": "\r",
        "e": "\x1b",
        "t": "\t",
        "@": "",
    }

    # pylint fails to realise ESCAPE_LOOKUP is defined
    # pylint: disable-msg = E0602

    ## @brief Mapping from unescaped string to corresponding escape character.
    REVERSE_ESCAPE_LOOKUP = dict((v, k) for k, v in ESCAPE_LOOKUP.items())

    # pylint: enable-msg = E0602

    ## @brief Regular expression matching all unescaped character.
    ESCAPE_RE = re.compile(r"[\\ \0\n\r\x1b\t]")

    ## @var mtype
    # @brief Message type.

    ## @var name
    # @brief Message name.

    ## @var arguments
    # @brief List of string message arguments.

    ## @brief Attempt to optimize messages by specifying attributes up front
    __slots__ = ["mtype", "name", "mid", "arguments"]

    def __init__(self, mtype, name, arguments=None, mid=None):
        self.mtype = mtype
        self.name = name

        if mid is None:
            self.mid = None
        else:
            self.mid = str(mid)

        if arguments is None:
            self.arguments = []
        else:
            self.arguments = [type(x) is float and repr(x) or str(x)
                              for x in arguments]

        # check message type

        if mtype not in self.TYPE_SYMBOLS:
            raise KatcpSyntaxError("Invalid command type %r." % (mtype,))

        # check message id

        if self.mid is not None and not self.mid.isdigit():
            raise KatcpSyntaxError("Invalid message id %r." % (mid,))

        # check command name validity

        if not name:
            raise KatcpSyntaxError("Command missing command name.")
        if not name.replace("-", "").isalnum():
            raise KatcpSyntaxError("Command name should consist only of"
                                " alphanumeric characters and dashes (got %r)."
                                % (name,))
        if not name[0].isalpha():
            raise KatcpSyntaxError("Command name should start with an"
                                " alphabetic character (got %r)."
                                % (name,))

    def copy(self):
        """Return a shallow copy of the message object and its arguments.

        Returns
        -------
        msg : Message
            A copy of the message object.
        """
        return Message(self.mtype, self.name, self.arguments)

    def __str__(self):
        """ Return Message serialized for transmission.

        Returns
        -------
        msg : str
           The message encoded as a ASCII string.
        """
        if self.arguments:
            escaped_args = [self.ESCAPE_RE.sub(self._escape_match, x)
                            for x in self.arguments]
            escaped_args = [x or "\\@" for x in escaped_args]
            arg_str = " " + " ".join(escaped_args)
        else:
            arg_str = ""

        if self.mid is not None:
            mid_str = "[%s]" % self.mid
        else:
            mid_str = ""

        return "%s%s%s%s" % (self.TYPE_SYMBOLS[self.mtype], self.name,
                             mid_str, arg_str)

    def __repr__(self):
        """ Return message displayed in a readable form
        """
        tp = self.TYPE_NAMES[self.mtype].lower()
        name = self.name
        if self.arguments:
            escaped_args = [self.ESCAPE_RE.sub(self._escape_match, x)
                            for x in self.arguments]
            for arg in escaped_args:
                if len(arg) > 10:
                    arg = arg[:10] + "..."
            args = "(" + ", ".join(escaped_args) + ")"
        else:
            args = ""
        return "<Message %s %s %s>" % (tp, name, args)

    def __eq__(self, other):
        if not isinstance(other, Message):
            return NotImplemented
        for name in self.__slots__:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    def __ne__(self, other):
        return not self == other

    def _escape_match(self, match):
        """Given a re.Match object, return the escape code for it."""
        return "\\" + self.REVERSE_ESCAPE_LOOKUP[match.group()]

    def reply_ok(self):
        """Return True if this is a reply and its first argument is 'ok'."""
        return (self.mtype == self.REPLY and self.arguments and
                self.arguments[0] == self.OK)

    # * and ** magic useful here
    # pylint: disable-msg = W0142

    @classmethod
    def request(cls, name, *args, **kwargs):
        """Helper method for creating request messages.

        Parameters
        ----------
        name : str
            The name of the message.
        args : list of strings
            The message arguments.
        """
        mid = kwargs.pop('mid', None)
        if len(kwargs) > 0:
            raise TypeError('Invalid keyword argument(s): %r' % kwargs)
        return cls(cls.REQUEST, name, args, mid)

    @classmethod
    def reply(cls, name, *args, **kwargs):
        """Helper method for creating reply messages.

        Parameters
        ----------
        name : str
            The name of the message.
        args : list of strings
            The message arguments.
        """
        mid = kwargs.pop('mid', None)
        if len(kwargs) > 0:
            raise TypeError('Invalid keyword argument(s): %r' % kwargs)
        return cls(cls.REPLY, name, args, mid)

    @classmethod
    def reply_to_request(cls, req_msg, *args):
        """Helper method for creating reply messages to a specific request.

        Copies the message name and message identifier from the request message

        Parameters
        ----------
        req_msg : katcp.core.Message instance
            The request message that this inform if in reply to
        args : list of strings
            The message arguments.
        """
        return cls(cls.REPLY, req_msg.name, args, req_msg.mid)


    @classmethod
    def inform(cls, name, *args, **kwargs):
        """Helper method for creating inform messages.

        Parameters
        ----------
        name : str
            The name of the message.
        args : list of strings
            The message arguments.
        """
        mid = kwargs.pop('mid', None)
        if len(kwargs) > 0:
            raise TypeError('Invalid keyword argument(s): %r' % kwargs)
        return cls(cls.INFORM, name, args, mid)

    @classmethod
    def reply_inform(cls, req_msg, *args):
        """Helper method for creating inform messages in reply to a request.

        Copies the message name and message identifier from the request message

        Parameters
        ----------
        req_msg : katcp.core.Message instance
            The request message that this inform if in reply to
        args : list of strings
            The message arguments except name
        """
        return cls(cls.INFORM, req_msg.name, args, req_msg.mid)

    # pylint: enable-msg = W0142


class MessageParser(object):
    """Parses lines into Message objects."""

    # We only want one public method
    # pylint: disable-msg = R0903

    ## @brief Copy of TYPE_SYMBOL_LOOKUP from Message.
    TYPE_SYMBOL_LOOKUP = Message.TYPE_SYMBOL_LOOKUP

    ## @brief Copy of ESCAPE_LOOKUP from Message.
    ESCAPE_LOOKUP = Message.ESCAPE_LOOKUP

    ## @brief Regular expression matching all special characters.
    SPECIAL_RE = re.compile(r"[\0\n\r\x1b\t ]")

    ## @brief Regular expression matching all escapes.
    UNESCAPE_RE = re.compile(r"\\(.?)")

    ## @brief Regular expresion matching KATCP whitespace (just space and tab)
    WHITESPACE_RE = re.compile(r"[ \t]+")

    ## @brief Regular expression matching name and ID
    NAME_RE = re.compile(
        r"^(?P<name>[a-zA-Z][a-zA-Z0-9\-]*)(\[(?P<id>[0-9]+)\])?$")

    def _unescape_match(self, match):
        """Given an re.Match, unescape the escape code it represents."""
        char = match.group(1)
        if char in self.ESCAPE_LOOKUP:
            return self.ESCAPE_LOOKUP[char]
        elif not char:
            raise KatcpSyntaxError("Escape slash at end of argument.")
        else:
            raise KatcpSyntaxError("Invalid escape character %r." % (char,))

    def _parse_arg(self, arg):
        """Parse an argument."""
        match = self.SPECIAL_RE.search(arg)
        if match:
            raise KatcpSyntaxError("Unescaped special %r." % (match.group(),))
        return self.UNESCAPE_RE.sub(self._unescape_match, arg)

    def parse(self, line):
        """Parse a line, return a Message.

        Parameters
        ----------
        line : str
            The line to parse (should not contain the terminating newline
            or carriage return).

        Returns
        -------
        msg : Message object
            The resulting Message.
        """
        # find command type and check validity
        if not line:
            raise KatcpSyntaxError("Empty message received.")

        type_char = line[0]
        if type_char not in self.TYPE_SYMBOL_LOOKUP:
            raise KatcpSyntaxError("Bad type character %r." % (type_char,))

        mtype = self.TYPE_SYMBOL_LOOKUP[type_char]

        # find command and arguments name
        # (removing possible empty argument resulting from whitespace at end
        #  of command)
        parts = self.WHITESPACE_RE.split(line)
        if not parts[-1]:
            del parts[-1]

        name = parts[0][1:]
        arguments = [self._parse_arg(x) for x in parts[1:]]

        # split out message id
        match = self.NAME_RE.match(name)
        if match:
            name = match.group('name')
            mid = match.group('id')
        else:
            raise KatcpSyntaxError("Bad message name (and possibly id) %r." %
                                   (name,))

        return Message(mtype, name, arguments, mid)


class ProtocolFlags(object):
    """Utility class for handling KATCP protocol flags.

    .. note::

       This class was introduced in katcp version 0.4.

    Currently understood flags are:

    * M - server supports multiple clients
    * I - server supports message identifiers

    Parameters
    ----------
    major : int
        Major version number.
    minor : int
        Minor version number.
    flags : set
        Set of supported flags.

    Attributes
    ----------
    multi_client : bool
        Whether the server the version string came from supports
        multiple clients.
    message_ids : bool
        Whether the server the version string came from supports
        message ids.
    """

    VERSION_RE = re.compile(r"^(?P<major>\d+)\.(?P<minor>\d+)"
                            r"(-(?P<flags>.*))?$")

    # flags

    MULTI_CLIENT = 'M'
    MESSAGE_IDS = 'I'

    STRATEGIES_V4 = frozenset(['none', 'auto', 'period', 'event', 'differential'])
    STRATEGIES_V5 = STRATEGIES_V4 | frozenset(
        ['event-rate', 'differential-rate'])

    STRATEGIES_ALLOWED_BY_MAJOR_VERSION = {
        4: STRATEGIES_V4,
        5: STRATEGIES_V5
        }

    def __init__(self, major, minor, flags):
        self.major = major
        self.minor = minor
        self.flags = set(list(flags))
        self.multi_client = self.MULTI_CLIENT in self.flags
        self.message_ids = self.MESSAGE_IDS in self.flags
        if self.message_ids and self.major < MID_KATCP_MAJOR:
            raise ValueError(
                'MESSAGE_IDS is only supported in katcp v5 and newer')

    def strategy_allowed(self, strategy):
        return strategy in self.STRATEGIES_ALLOWED_BY_MAJOR_VERSION[self.major]

    def __eq__(self, other):
        if not isinstance(other, ProtocolFlags):
            return NotImplemented
        return (self.major == other.major and self.minor == other.minor
                and self.flags == other.flags)

    def __str__(self):
        flag_str = self.flags and ("-" + "".join(sorted(self.flags))) or ""
        return "%d.%d%s" % (self.major, self.minor, flag_str)

    def supports(self, flag):
        return flag in self.flags

    @classmethod
    def parse_version(cls, version_str):
        """Create a :class:`ProtocolFlags` object from a version string.

        Parameters
        ----------
        version_str : str
            The version string from a #version-connect katcp-protocol
            message.
        """
        match = cls.VERSION_RE.match(version_str)
        if match:
            major = int(match.group('major'))
            minor = int(match.group('minor'))
            flags = set(match.group('flags') or '')
        else:
            major, minor, flags = None, None, set()
        return cls(major, minor, flags)


class DeviceMetaclass(type):
    """Metaclass for DeviceServer and DeviceClient classes.

       Collects up methods named request\_* and adds
       them to a dictionary of supported methods on the class.
       All request\_* methods must have a doc string so that help
       can be generated.  The same is done for inform\_* and
       reply\_* methods.
       """

    def __init__(mcs, name, bases, dct):
        """Constructor for DeviceMetaclass.  Should not be used directly.

        Parameters
        ----------
        mcs : class
            The metaclass instance
        name : str
            The metaclass name
        bases : list of classes
            List of base classes
        dct : dict
            Class dictionary
        """
        super(DeviceMetaclass, mcs).__init__(name, bases, dct)
        mcs._request_handlers = {}
        mcs._inform_handlers = {}
        mcs._reply_handlers = {}

        def convert(prefix, name):
            """Convert a method name to the corresponding command name."""
            return name[len(prefix):].replace("_", "-")

        for name in dir(mcs):
            if not callable(getattr(mcs, name)):
                continue
            if name.startswith("request_"):
                request_name = convert("request_", name)
                mcs._request_handlers[request_name] = getattr(mcs, name)
                assert(mcs._request_handlers[request_name].__doc__ is not None)
            elif name.startswith("inform_"):
                inform_name = convert("inform_", name)
                mcs._inform_handlers[inform_name] = getattr(mcs, name)
                assert(mcs._inform_handlers[inform_name].__doc__ is not None)
                # There is a bit of a name colission between the reply_*
                # convention and the server reply_inform() method
            elif name.startswith("reply_") and name != 'reply_inform':
                reply_name = convert("reply_", name)
                mcs._reply_handlers[reply_name] = getattr(mcs, name)
                assert(mcs._reply_handlers[reply_name].__doc__ is not None)


class KatcpDeviceError(Exception):
    """Raised by KATCP servers when errors occur.

    .. versionchanged:: 0.1
        Deprecated in 0.1. Servers should not raise errors if communication
        with a client fails -- errors are simply logged instead.
    """
    pass


class FailReply(Exception):
    """Raised by request handlers to indicate a failure.

    A custom exception which, when thrown in a request handler,
    causes DeviceServerBase to send a fail reply with the specified
    fail message, bypassing the generic exception handling, which
    would send a fail reply with a full traceback.

    Examples
    --------
    >>> class MyDevice(DeviceServer):
    ...     def request_myreq(self, req, msg):
    ...         raise FailReply("This request always fails.")
    ...
    """
    pass


class AsyncReply(Exception):
    """Raised by a request handlers to indicate it will reply later.

    A custom exception which, when thrown in a request handler,
    indicates to DeviceServerBase that no reply has been returned
    by the handler but that the handler has arranged for a reply
    message to be sent at a later time.

    Examples
    --------
    >>> class MyDevice(DeviceServer):
    ...     def request_myreq(self, req, msg):
    ...         self.callback_client.request(
    ...             Message.request("otherreq"),
    ...             reply_cb=self._send_reply,
    ...         )
    ...         raise AsyncReply()
    ...

    """
    pass


class ExcepthookThread(threading.Thread):
    """A custom Thread class that provides an exception hook.

    Exceptions are passed up to an excepthook callable that
    functions like sys.excepthook.

    Parameters
    ----------
    excepthook : callable
        Function to call when the thread raises an unhandled
        exception. The signature is the same as for sys.excepthook.
    args : additional arguments
        Passed to the threading.Thread constructor.
    kwargs: additional keyword arguments
        Passed to the threading.Thread constructor.
    """
    def __init__(self, excepthook=None, *args, **kwargs):
        if excepthook is None:
            excepthook = getattr(threading.currentThread(), "_excepthook",
                                 None)
        self._excepthook = excepthook
        # evil hack to support subclasses that override run
        self._old_run = self.run
        self.run = self._wrapped_run
        super(ExcepthookThread, self).__init__(*args, **kwargs)

    def _wrapped_run(self):
        try:
            self._old_run()
        except:
            if self._excepthook is not None:
                self._excepthook(*sys.exc_info())
            else:
                raise


from .kattypes import Int, Float, Bool, Discrete, Lru, Str, Timestamp, Address


class Sensor(object):
    """Instantiate a new sensor object.

    Subclasses will usually pass in a fixed sensor_type which should
    be one of the sensor type constants. The list params if set will
    have its values formatter by the type formatter for the given
    sensor type.

    .. note::

       The LRU sensor type was deprecated in katcp 0.4.

    .. note::

       The ADDRESS sensor type was added in katcp 0.4.

    Parameters
    ----------
    sensor_type : Sensor type constant
        The type of sensor.
    name : str
        The name of the sensor.
    description : str
        A short description of the sensor.
    units : str
        The units of the sensor value. May be the empty string
        if there are no applicable units.
    params : list
        Additional parameters, dependent on the type of sensor:

          * For :const:`INTEGER` and :const:`FLOAT` the list should
            give the minimum and maximum that define the range
            of the sensor value.
          * For :const:`DISCRETE` the list should contain all
            possible values the sensor may take.
          * For all other types, params should be omitted.
    default : object
        An initial value for the sensor. By default this is
        determined by the sensor type.
    """

    # Sensor needs the instance attributes it has and
    # is an abstract class used only outside this module
    # pylint: disable-msg = R0902

    # Type names and formatters
    #
    # Formatters take the sensor object and the value to
    # be formatted as arguments. They may raise exceptions
    # if the value cannot be formatted.
    #
    # Parsers take the sensor object and the value to
    # parse as arguments
    #
    # type -> (name, formatter, parser)
    (INTEGER, FLOAT, BOOLEAN, LRU, DISCRETE, STRING, TIMESTAMP,
     ADDRESS) = range(8)

    ## @brief Mapping from sensor type to tuple containing the type name,
    #  a kattype with functions to format and parse a value and a
    #  default value for sensors of that type.
    SENSOR_TYPES = {
        INTEGER: (Int, 0),
        FLOAT: (Float, 0.0),
        BOOLEAN: (Bool, False),
        LRU: (Lru, Lru.LRU_NOMINAL),
        DISCRETE: (Discrete, "unknown"),
        STRING: (Str, ""),
        TIMESTAMP: (Timestamp, 0.0),
        ADDRESS: (Address, Address.NULL),
    }

    SENSOR_SHORTCUTS = {
        int: INTEGER,
        float: FLOAT,
        str: STRING,
        bool: BOOLEAN,
      }

    # map type strings to types
    SENSOR_TYPE_LOOKUP = dict((v[0].name, k) for k, v in SENSOR_TYPES.items())

    # Sensor status constants
    UNKNOWN, NOMINAL, WARN, ERROR, FAILURE, UNREACHABLE, INACTIVE = range(7)

    ## @brief Mapping from sensor status to status name.
    STATUSES = {
        UNKNOWN: 'unknown',
        NOMINAL: 'nominal',
        WARN: 'warn',
        ERROR: 'error',
        FAILURE: 'failure',
        UNREACHABLE: 'unreachable',
        INACTIVE: 'inactive',
    }

    ## @brief Mapping from status name to sensor status.
    STATUS_NAMES = dict((v, k) for k, v in STATUSES.items())

    # LRU sensor values
    LRU_NOMINAL, LRU_ERROR = Lru.LRU_NOMINAL, Lru.LRU_ERROR

    ## @brief Mapping from LRU value constant to LRU value name.
    LRU_VALUES = Lru.LRU_VALUES

    # LRU_VALUES not found by pylint
    # pylint: disable-msg = E0602

    ## @brief Mapping from LRU value name to LRU value constant.
    LRU_CONSTANTS = dict((v, k) for k, v in LRU_VALUES.items())

    # pylint: enable-msg = E0602

    ## @brief kattype Timestamp instance for encoding and decoding timestamps
    TIMESTAMP_TYPE = Timestamp()

    ## @var stype
    # @brief Sensor type constant.

    ## @var name
    # @brief Sensor name.

    ## @var description
    # @brief String describing the sensor.

    ## @var units
    # @brief String contain the units for the sensor value.

    ## @var params
    # @brief List of strings containing the additional parameters (length and
    #        interpretation are specific to the sensor type)

    def __init__(self, sensor_type, name, description=None, units='', params=None,
                 default=None):
        if params is None:
            params = []

        sensor_type = self.SENSOR_SHORTCUTS.get(sensor_type, sensor_type)

        self._sensor_type = sensor_type
        self._observers = set()

        typeclass, default_value = self.SENSOR_TYPES[sensor_type]

        if self._sensor_type in [Sensor.INTEGER, Sensor.FLOAT]:
            # as of version 5 of the guidelines, integer and float
            # ranges are optional and informational
            if len(params) == 2:
                if not params[0] <= default_value <= params[1]:
                    default_value = params[0]
            self._kattype = typeclass()
        elif self._sensor_type == Sensor.DISCRETE:
            default_value = params[0]
            self._kattype = typeclass(params)
        else:
            if self._sensor_type == Sensor.TIMESTAMP and units:
                raise ValueError(
                    'Units cannot be specified for TIMESTAMP sensors since '
                    'their units is defined by the KATCP spec as either '
                    'seconds or, for katcp versions 4 and below, milliseconds')
            self._kattype = typeclass()

        if default is not None:
            default_value = default

        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # self._value_tuple should also be set and read in a single
        # bytecode to avoid situations were an update in one thread
        # causes another thread to read the timestamp from one update
        # and the value and/or status from a different update.
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        self._value_tuple = (time.time(), Sensor.UNKNOWN, default_value)
        self._formatter = self._kattype.pack
        self._parser = self._kattype.unpack
        self.stype = self._kattype.name

        self.name = name
        if description is None:
            description = '%(type)s sensor %(name)r %(unit_description)s' % dict(
               type=self.stype.capitalize(), name=self.name,
               unit_description=('in unit '+units if units else 'with no unit'))

        self.description = description
        self.units = units
        self.params = params
        self.formatted_params = [self._formatter(p, True) for p in params]

    # support for legacy KATCP users that relied on being able to
    # read _timestamp, _status and _value. Such usage will be
    # deprecated in a future version of KATCP.

    def _value_tuple_getter(i, name):
        def getter(self):
            warnings.warn("Use of katcp.Sensor.%s attribute is deprecated"
                          % name, DeprecationWarning)
            return self._value_tuple[i]
        return getter

    _timestamp = property(_value_tuple_getter(0, "_timestamp"))
    _status = property(_value_tuple_getter(1, "_status"))
    _value = property(_value_tuple_getter(2, "_value"))

    del _value_tuple_getter

    def __repr__(self):
        cls = self.__class__
        return "<%s.%s object name=%r at 0x%x>" % \
               (cls.__module__, cls.__name__, self.name, id(self))

    @classmethod
    def integer(cls, name, description=None, unit='', params=None, default=None):
        """
        Instantiate a new integer sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units.
        params : list
            [min, max] -- miniumum and maximum values of the sensor
        default : int
            An initial value for the sensor. Defaults to 0.
        """
        return cls(cls.INTEGER, name, description, unit, params, default)

    @classmethod
    def float(cls, name, description=None, unit='', params=None, default=None):
        """
        Instantiate a new float sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units.
        params : list
            [min, max] -- miniumum and maximum values of the sensor
        default : float
            An initial value for the sensor. Defaults to 0.0.
        """
        return cls(cls.FLOAT, name, description, unit, params, default)

    @classmethod
    def boolean(cls, name, description=None, unit='', default=None):
        """
        Instantiate a new boolean sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units.
        default : bool
            An initial value for the sensor. Defaults to False.
        """
        return cls(cls.BOOLEAN, name, description, unit, None, default)

    @classmethod
    def lru(cls, name, description=None, unit='', default=None):
        """
        Instantiate a new lru sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units.
        default : enum, Sensor.LRU_*
            An initial value for the sensor. Defaults to self.LRU_NOMINAL
        """
        return cls(cls.LRU, name, description, unit, None, default)

    @classmethod
    def string(cls, name, description=None, unit='', default=None):
        """
        Instantiate a new string sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units.
        default : string
            An initial value for the sensor. Defaults to the empty string.
        """
        return cls(cls.STRING, name, description, unit, None, default)

    @classmethod
    def discrete(cls, name, description=None, unit='', params=None, default=None):
        """
        Instantiate a new discrete sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units.
        params : [str]
            Sequence of all allowable discrete sensor states
        default : str
            An initial value for the sensor. Defaults to the first item
            of params
        """
        return cls(cls.DISCRETE, name, description, unit, params, default)

    @classmethod
    def timestamp(cls, name, description=None, unit='seconds', default=None):
        """
        Instantiate a new timestamp sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units. Defaults to 'seconds'.
        default : string
            An initial value for the sensor in seconds since the Unix Epoch.
            Defaults to 0.
        """
        return cls(cls.TIMESTAMP, name, description, unit, None, default)

    @classmethod
    def address(cls, name, description=None, unit='', default=None):
        """
        Instantiate a new IP address sensor object.

        name : str
            The name of the sensor.
        description : str
            A short description of the sensor.
        units : str
            The units of the sensor value. May be the empty string
            if there are no applicable units.
        default : (string, int)
            An initial value for the sensor. Tuple contaning (host, port).
            default is ("0.0.0.0", None)
        """
        return cls(cls.ADDRESS, name, description, unit, None, default)


    def attach(self, observer):
        """Attach an observer to this sensor.

        The observer must support a call to observer.update(sensor).

        Parameters
        ----------
        observer : object
            Object with an .update(sensor) method that will be called
            when the sensor value is set.
        """
        self._observers.add(observer)

    def detach(self, observer):
        """Detach an observer from this sensor.

        Parameters
        ----------
        observer : object
            The observer to remove from the set of observers notified
            when the sensor value is set.
        """
        self._observers.discard(observer)

    def notify(self):
        """Notify all observers of changes to this sensor."""
        # copy list before iterating in case new observers arrive
        for o in list(self._observers):
            o.update(self)

    def parse_value(self, s_value, katcp_major=DEFAULT_KATCP_MAJOR):
        """Parse a value from a string.

        Parameters
        ----------
        s_value : str
            A string value to attempt to convert to a value for
            the sensor.

        Returns
        -------
        value : object
            A value of a type appropriate to the sensor.
        """
        return self._parser(s_value, katcp_major)

    def set(self, timestamp, status, value):
        """Set the current value of the sensor.

        Parameters
        ----------
        timestamp : float in seconds
           The time at which the sensor value was determined.
        status : Sensor status constant
            Whether the value represents an error condition or not.
        value : object
            The value of the sensor (the type should be appropriate to the
            sensor's type).
        """
        self._value_tuple = (timestamp, status, value)
        self.notify()

    def set_formatted(self, raw_timestamp, raw_status, raw_value,
                      major=DEFAULT_KATCP_MAJOR):
        """Set the current value of the sensor.

        Parameters
        ----------
        timestamp : str
            KATCP formatted timestamp string
        status : str
            KATCP formatted sensor status string
        value : str
            KATCP formatted sensor value
        major : int, default = 5
            KATCP major version to use for interpreting the raw values
        """
        timestamp = self.TIMESTAMP_TYPE.decode(raw_timestamp, major)
        status = self.STATUS_NAMES[raw_status]
        value = self.parse_value(raw_value, major)
        self.set(timestamp, status, value)

    def read_formatted(self, major=DEFAULT_KATCP_MAJOR):
        """Read the sensor and return a timestamp, status, value tuple.

        All values are strings formatted as specified in the Sensor Type
        Formats in the katcp specification.

        Parameters
        ----------
        major : int. Defaults to latest implemented KATCP version (5)
            Major version of KATCP to use when interpreting types

        Returns
        -------
        timestamp : str
            KATCP formatted timestamp string
        status : str
            KATCP formatted sensor status string
        value : str
            KATCP formatted sensor value
        """
        timestamp, status, value = self.read()
        return (self.TIMESTAMP_TYPE.encode(timestamp, major),
                self.STATUSES[status],
                self._formatter(value, True, major))

    def read(self):
        """Read the sensor and return a timestamp, status, value tuple.

        Returns
        -------
        timestamp : float in seconds
           The time at which the sensor value was determined.
        status : Sensor status constant
            Whether the value represents an error condition or not.
        value : object
            The value of the sensor (the type will be appropriate to the
            sensor's type).
        """
        return self._value_tuple

    def set_value(self, value, status=NOMINAL, timestamp=None,
                  major=DEFAULT_KATCP_MAJOR):
        """Check and then set the value of the sensor.

        Parameters
        ----------
        value : object
            Value of the appropriate type for the sensor.
        status : Sensor status constant
            Whether the value represents an error condition or not.
        timestamp : float in seconds or None
           The time at which the sensor value was determined. Uses current time
           if None.
        major : int. Defaults to latest implemented KATCP version (5)
            Major version of KATCP to use when interpreting types
        """
        self._kattype.check(value, major)
        if timestamp is None:
            timestamp = time.time()
        self.set(timestamp, status, value)

    def value(self):
        """Read the current sensor value.

        Returns
        -------
        value : object
            The value of the sensor (the type will be appropriate to the
            sensor's type).
        """
        return self.read()[2]

    @classmethod
    def parse_type(cls, type_string):
        """Parse KATCP formatted type code into Sensor type constant.

        Parameters
        ----------
        type_string : str
            KATCP formatted type code.

        Returns
        -------
        sensor_type : Sensor type constant
            The corresponding Sensor type constant.
        """
        if type_string in cls.SENSOR_TYPE_LOOKUP:
            return cls.SENSOR_TYPE_LOOKUP[type_string]
        else:
            raise KatcpSyntaxError("Invalid sensor type string %s" %
                                   type_string)

    @classmethod
    def parse_params(cls, sensor_type, formatted_params,
                     major=DEFAULT_KATCP_MAJOR):
        """Parse KATCP formatted parameters into Python values.

        Parameters
        ----------
        sensor_type : Sensor type constant
            The type of sensor the parameters are for.
        formatted_params : list of strings
            The formatted parameters that should be parsed.
        major : int. Defaults to latest implemented KATCP version (5)
            Major version of KATCP to use when interpreting types

        Returns
        -------
        params : list of objects
            The parsed parameters.
        """
        typeclass, _value = cls.SENSOR_TYPES[sensor_type]
        if sensor_type == cls.DISCRETE:
            kattype = typeclass([])
        else:
            kattype = typeclass()
        return [kattype.decode(x, major) for x in formatted_params]



# __init__.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2010 SKA South Africa (http://ska.ac.za/)
# BSD license - see COPYING for details

"""Tests for the katcp.tx package.
   """
### Horrendious monkey patch hack to make twisted.trial.unittest
### support skipping in python 2.6
import unittest, unittest2
oldTestCase = unittest.TestCase
unittest.TestCase = unittest2.TestCase

from twisted.trial import unittest as tx_unit
unittest.TestCase = oldTestCase

def addSkip(self, test, reason):
    self.original.addSkip(test, reason)

tx_unit.PyUnitResultAdapter.addSkip = addSkip
### End horrendious hack

def suite():
    import unittest
    import test_core
    import test_proxy
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTests(loader.loadTestsFromModule(test_core))
    suite.addTests(loader.loadTestsFromModule(test_proxy))
    return suite



"""Tests for the nodeman module."""

import unittest
import katcp


class BaseTreeTest(unittest.TestCase):

    def assertSensorValues(self, sensors, values):
        for sensor, value in zip(sensors, values):
            self.assertEqual(sensor.value(), value, "Expected %r to equal %s" %
                             (sensor, value))

    @staticmethod
    def make_sensors(number, stype, params=None):
        """Create a number of sensors of the same type."""
        sensors = []
        for i in range(number):
            sensors.append(katcp.Sensor(stype, "sensor%d" % i,
                "Boolean Sensor %d" % i, "", params=params))
        return sensors


class RecordingTree(katcp.GenericSensorTree):

    def __init__(self):
        super(RecordingTree, self).__init__()
        self.calls = []

    def recalculate(self, parent, updates):
        self.calls.append((parent, updates))
        # dummy update of the parent sensor value
        parent.set_value(parent.value())


class TestGenericSensorTree(BaseTreeTest):

    def setUp(self):
        self.tree = RecordingTree()
        self.calls = self.tree.calls
        self.sensor1 = katcp.Sensor(int, "sensor1", "First sensor", "",
                                    [0, 100])
        self.sensor2 = katcp.Sensor(int, "sensor2", "Second sensor", "",
                                    [0, 100])
        self.sensor3 = katcp.Sensor(int, "sensor3", "Third sensor", "",
                                    [0, 100])

    def test_add_links(self):
        self.tree.add_links(self.sensor1, [self.sensor2])
        self.assertEqual(self.calls, [(self.sensor1, [self.sensor2])])
        self.assertEqual(self.sensor1._observers, set([self.tree]))
        self.assertEqual(self.sensor2._observers, set([self.tree]))
        self.assertEqual(self.tree.children(self.sensor1), set([self.sensor2]))
        self.assertEqual(self.tree.children(self.sensor2), set())
        self.assertEqual(self.tree.parents(self.sensor1), set())
        self.assertEqual(self.tree.parents(self.sensor2), set([self.sensor1]))

    def test_remove_links(self):
        self.tree.add_links(self.sensor1, [self.sensor2])
        self.tree.remove_links(self.sensor1, [self.sensor2])
        self.assertEqual(self.calls, [
            (self.sensor1, [self.sensor2]),
            (self.sensor1, [self.sensor2]),
        ])
        self.assertEqual(self.sensor1._observers, set())
        self.assertEqual(self.sensor2._observers, set())
        self.assertRaises(ValueError, self.tree.children, self.sensor1)
        self.assertRaises(ValueError, self.tree.children, self.sensor2)
        self.assertRaises(ValueError, self.tree.parents, self.sensor1)
        self.assertRaises(ValueError, self.tree.parents, self.sensor2)

    def test_concurrent_update(self):
        # this test adds a new parent to a sensor while the sensor
        # is being updated. If it fails the call to set_value at the
        # end will raise 'RuntimeError: Set changed size during iteration'.
        self.tree.add_links(self.sensor1, [self.sensor2])

        class LinkOnUpdate(object):
            def update(linker, sensor):
                self.tree.add_links(self.sensor3, [self.sensor2])

        linker = LinkOnUpdate()
        self.sensor1.attach(linker)
        self.sensor2.set_value(3)
        self.assertEqual(self.tree.parents(self.sensor2),
                         set([self.sensor1, self.sensor3]))


class TestBooleanSensorTree(BaseTreeTest):

    def test_basic(self):
        tree = katcp.BooleanSensorTree()
        s0, s1, s2, s3 = sensors = self.make_sensors(4, katcp.Sensor.BOOLEAN)
        tree.add(s0, s1)
        tree.add(s0, s2)
        tree.add(s1, s3)
        self.assertSensorValues(sensors, (False, False, False, False))

        s3.set_value(True)
        self.assertSensorValues(sensors, (False, True, False, True))
        s2.set_value(True)
        self.assertSensorValues(sensors, (True, True, True, True))
        s3.set_value(False)
        self.assertSensorValues(sensors, (False, False, True, False))
        s2.set_value(False)
        self.assertSensorValues(sensors, (False, False, False, False))

        s3.set_value(True)
        tree.remove(s0, s2)
        self.assertSensorValues(sensors, (True, True, False, True))

        s3.set_value(False)
        tree.remove(s1, s3)
        self.assertSensorValues(sensors, (True, True, False, False))

        s1.set_value(False)
        tree.remove(s0, s1)
        self.assertSensorValues(sensors, (True, False, False, False))


class TestAggregateSensorTree(BaseTreeTest):

    @staticmethod
    def _add_rule(parent, children):
        """Set a parent sensor to the sum of its children."""
        parent.set_value(sum(child.value() for child in children))

    def test_basic(self):
        tree = katcp.AggregateSensorTree()
        s0, s1, s2, s3 = sensors = self.make_sensors(4, katcp.Sensor.INTEGER,
                                                     params=[-100, 100])
        tree.add(s0, self._add_rule, (s1,))
        tree.add(s1, self._add_rule, (s2, s3))
        self.assertSensorValues(sensors, (0, 0, 0, 0))

        s2.set_value(1)
        self.assertSensorValues(sensors, (1, 1, 1, 0))
        s3.set_value(2)
        self.assertSensorValues(sensors, (3, 3, 1, 2))

        tree.remove(s1)
        self.assertSensorValues(sensors, (3, 3, 1, 2))

        s1.set_value(5)
        self.assertSensorValues(sensors, (5, 5, 1, 2))

        tree.remove(s0)
        self.assertSensorValues(sensors, (5, 5, 1, 2))

        # check s0 is no longer updated
        s1.set_value(7)
        self.assertSensorValues(sensors, (5, 7, 1, 2))

    def test_double_add(self):
        tree = katcp.AggregateSensorTree()
        s0, s1 = self.make_sensors(2, katcp.Sensor.INTEGER, params=[-100, 100])
        tree.add(s0, self._add_rule, (s1,))
        self.assertRaises(ValueError, tree.add, s0, self._add_rule, (s1,))

    def test_double_remove(self):
        tree = katcp.AggregateSensorTree()
        s0, s1 = self.make_sensors(2, katcp.Sensor.INTEGER, params=[-100, 100])
        tree.add(s0, self._add_rule, (s1,))
        tree.remove(s0)
        self.assertRaises(ValueError, tree.remove, s0)

    def test_delayed(self):
        tree = katcp.AggregateSensorTree()
        s0, s1, s2, s3 = sensors = self.make_sensors(4, katcp.Sensor.INTEGER,
                                                     params=[-100, 100])
        tree.add_delayed(s0, self._add_rule, (s1.name,))
        tree.add_delayed(s1, self._add_rule, (s2.name, s3.name))
        self.assertSensorValues(sensors, (0, 0, 0, 0))

        s2.set_value(1)
        s3.set_value(2)
        self.assertSensorValues(sensors, (0, 0, 1, 2))

        tree.register_sensor(s2)
        tree.register_sensor(s3)
        self.assertSensorValues(sensors, (0, 3, 1, 2))

        tree.register_sensor(s1)
        self.assertSensorValues(sensors, (3, 3, 1, 2))




from katcp.tx.core import DeviceServer, ClientKatCPProtocol, DeviceProtocol
from twisted.internet.defer import DeferredList
from twisted.internet import reactor
from twisted.internet.protocol import ClientFactory
from twisted.python import log
from katcp import Message, AsyncReply, Sensor
from katcp.kattypes import request, return_reply, Int

import re
import time


def value_only_formatted(func):
    """ A decorator that changes a value-only read into read_formatted
    format (using time.time and 'ok')
    """
    def new_func(self):
        return time.time(), "ok", func(self)
    new_func.func_name = func.func_name
    return new_func


class ProxiedSensor(Sensor):
    """ A sensor which is a proxy for other sensor on the remote device.
    Returns a deferred on read
    """
    def __init__(self, name, description, units, stype, device, proxy,
                 *formatted_params):
        self.basename = name
        self.device = device
        stype = Sensor.parse_type(stype)
        params = Sensor.parse_params(stype, formatted_params)
        Sensor.__init__(self, stype, device.name + '.' + name, description,
                        units, params=params)

    def read_formatted(self):
        return self.device.send_request('sensor-value', self.basename)


class StateSensor(object):
    """ A device state sensor
    """
    description = 'connection state'
    stype = 'discrete'
    formatted_params = ('unsynced', 'syncing', 'synced')
    units = ''

    def __init__(self, name, device):
        self.device = device
        self.name = name

    @value_only_formatted
    def read_formatted(self):
        return DeviceHandler.STATE_NAMES[self.device.state]


class DeviceHandler(ClientKatCPProtocol):
    SYNCING, SYNCED, UNSYNCED = range(3)
    STATE_NAMES = ['syncing', 'synced', 'unsynced']

    TYPE = 'full'

    stopping = False

    _conn_counter = 0

    SensorClass = ProxiedSensor

    def __init__(self, name, host, port):
        self.name = name
        self.host = host
        self.port = port
        ClientKatCPProtocol.__init__(self)
        self.requests = []
        self.sensors = {}
        self.state = self.UNSYNCED

    def _got_help(self, (informs, reply)):
        for inform in informs:
            self.requests.append(inform.arguments[0])
        self.send_request('sensor-list').addCallback(self._got_sensor_list)

    def _got_sensor_list(self, (informs, reply)):
        self.state = self.SYNCED
        for inform in informs:
            name, description, units, stype = inform.arguments[:4]
            formatted_arguments = inform.arguments[4:]
            sensor = self.SensorClass(name, description,
                                      units, stype,
                                      self, self.proxy, *formatted_arguments)
            self.sensors[name] = sensor
            self.proxy.add_proxied_sensor(self, sensor)
        self.device_ready()
        self.proxy.device_ready(self)

    def device_ready(self):
        """ Another hook that can be overloaded if you want to execute
        code just after device has been synced
        """
        pass

    def connectionMade(self):
        """ This is called after connection has been made. Introspect server
        about it's capabilities
        """
        self.state = self.SYNCING
        self.send_request('help').addCallback(self._got_help)
        self._conn_counter = 0

    def add_proxy(self, proxy):
        self.proxy = proxy
        proxy.add_sensor(StateSensor(self.name + '-' + 'state', self))

    def schedule_resyncing(self):
        reactor.connectTCP(self.host, self.port, self.proxy.client_factory)

    def connectionLost(self, failure):
        self.state = self.UNSYNCED
        ClientKatCPProtocol.connectionLost(self, failure)
        if not self.stopping:
            reactor.callLater(self.proxy.CONN_DELAY_TIMEOUT,
                              self.schedule_resyncing)

    def stop(self):
        """ A hook for stopping requested device, if necessary, does nothing
        by default
        """
        pass

    def inform_sensor_status(self, msg):
        try:
            sensor = self.sensors[msg.arguments[2]]
            sensor.set_formatted(msg.arguments[0], msg.arguments[3],
                                 msg.arguments[4])
        except:
            log.err()


class ProxyProtocol(DeviceProtocol):
    @request(include_msg=True)
    @return_reply(Int(min=0))
    def request_device_list(self, reqmsg):
        """Return a list of devices aggregated by the proxy.

        Returns the list of devices a sequence of #device-list informs.

        Inform Arguments
        ----------------
        device : str
            Name of a device.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the list of devices succeeded.
        informs : int
            Number of #device-list informs sent.

        Examples
        --------
        ?device-list
        #device-list antenna
        #device-list enviro
        !device-list ok 2
        """
        for name in sorted(self.factory.devices):
            self.send_reply_inform(Message.inform("device-list", name,
                                   self.factory.devices[name].TYPE), reqmsg)
        return "ok", len(self.factory.devices)

    def request_sensor_list(self, msg):
        """Request the list of sensors.

        The list of sensors is sent as a sequence of #sensor-list informs.

        Parameters
        ----------
        name : str or pattern, optional
            If the name is not a pattern, list just the sensor with the given
            name. A pattern starts and ends with a slash ('/') and uses the
            Python re module's regular expression syntax. All sensors whose
            names contain the pattern are listed.  The default is to list all
            sensors.

        Inform Arguments
        ----------------
        name : str
            The name of the sensor being described.
        description : str
            Description of the named sensor.
        units : str
            Units for the value of the named sensor.
        type : str
            Type of the named sensor.
        params : list of str, optional
            Additional sensor parameters (type dependent). For integer and
            float sensors the additional parameters are the minimum and maximum
            sensor value. For discrete sensors the additional parameters are
            the allowed values. For all other types no additional parameters
            are sent.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the sensor list succeeded.
        informs : int
            Number of #sensor-list inform messages sent.

        Examples
        --------
        ?sensor-list
        #sensor-list psu.voltage PSU\_voltage. V float 0.0 5.0
        #sensor-list cpu.status CPU\_status. \@ discrete on off error
        ...
        !sensor-list ok 5

        ?sensor-list /voltage/
        #sensor-list psu.voltage PSU\_voltage. V float 0.0 5.0
        #sensor-list cpu.voltage CPU\_voltage. V float 0.0 3.0
        !sensor-list ok 2

        ?sensor-list cpu.power.on
        #sensor-list cpu.power.on Whether\_CPU\_hase\_power. \@ boolean
        !sensor-list ok 1
        """
        # handle non-regex cases
        if not msg.arguments or not (msg.arguments[0].startswith("/")
            and msg.arguments[0].endswith("/")):
            return DeviceProtocol.request_sensor_list(self, msg)

        # handle regex
        name_re = re.compile(msg.arguments[0][1:-1])
        sensors = dict([(name, sensor) for name, sensor in
            self.factory.sensors.iteritems() if name_re.search(name)])

        for name, sensor in sorted(sensors.items(), key=lambda x: x[0]):
            self.send_reply_inform(Message.inform("sensor-list",
                name, sensor.description, sensor.units, sensor.stype,
                *sensor.formatted_params), msg)

        return Message.reply(msg.name, "ok", len(sensors))

    def _send_all_sensors(self, reqmsg, filter=None):
        """ Sends all sensor values with given filter (None = all)
        """
        counter = [0]  # this has to be a list or an object, thanks to
        # python lexical scoping rules (we could not write count += 1
        # in a function)

        def device_ok((informs, reply), device):
            for inform in informs:
                inform.arguments[2] = device.name + '.' + \
                                      inform.arguments[2]
                if filter is None or re.match(filter, inform.arguments[2]):
                    self.send_reply_inform(inform, reqmsg)
                    counter[0] += 1

        def all_ok(_):
            self.send_reply(Message.reply('sensor-value', 'ok',
                                          str(counter[0])),
                            reqmsg)

        wait_for = []
        for device in self.factory.devices.itervalues():
            if device.state == device.SYNCED:
                d = device.send_request('sensor-value')
                d.addCallback(device_ok, device)
                wait_for.append(d)
            # otherwise we don't have the list of sensors, so we don't
            # send the message
        DeferredList(wait_for).addCallback(all_ok)
        for name, sensor in self.factory.sensors.iteritems():
            if not isinstance(sensor, ProxiedSensor):
                if filter is None or re.match(filter, name):
                    timestamp_ms, status, value = sensor.read_formatted()
                    counter[0] += 1
                    self.send_reply_inform(Message.inform('sensor-value',
                                                          timestamp_ms, "1",
                                                          name, status,
                                                          value),
                                           reqmsg)

    def request_sensor_value(self, msg):
        """Poll a sensor value or value(s).

        A list of sensor values as a sequence of #sensor-value informs.

        Parameters
        ----------
        name : str or pattern, optional
            If the name is not a pattern, list just the values of sensors with
            the given name.  A pattern starts and ends with a slash ('/') and
            uses the Python re module's regular expression syntax. The values
            of all sensors whose names contain the pattern are listed.  The
            default is to list the values of all sensors.

        Inform Arguments
        ----------------
        timestamp : float
            Timestamp of the sensor reading in milliseconds since the Unix
            epoch.
        count : {1}
            Number of sensors described in this #sensor-value inform. Will
            always be one. It exists to keep this inform compatible with
            #sensor-status.
        name : str
            Name of the sensor whose value is being reported.
        value : object
            Value of the named sensor. Type depends on the type of the sensor.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the list of values succeeded.
        informs : int
            Number of #sensor-value inform messages sent.

        Examples
        --------
        ?sensor-value
        #sensor-value 1244631611415.231 1 psu.voltage 4.5
        #sensor-value 1244631611415.200 1 cpu.status off
        ...
        !sensor-value ok 5

        ?sensor-value /voltage/
        #sensor-value 1244631611415.231 1 psu.voltage 4.5
        #sensor-value 1244631611415.100 1 cpu.voltage 4.5
        !sensor-value ok 2

        ?sensor-value cpu.power.on
        #sensor-value 1244631611415.231 1 cpu.power.on 0
        !sensor-value ok 1
        """
        if not msg.arguments:
            self._send_all_sensors(msg)
            raise AsyncReply()
        name = msg.arguments[0]
        if len(name) >= 2 and name.startswith("/") and name.endswith("/"):
            # regex case
            self._send_all_sensors(msg, name[1:-1])
            raise AsyncReply()
        else:
            return DeviceProtocol.request_sensor_value(self, msg)

    def request_halt(self, msg):
        """ drops connection to specified device
        """
        def got_halt((informs, reply)):
            self.send_reply(Message.reply('halt', dev_name, 'ok'), msg)

        if not msg.arguments:
            return DeviceProtocol.halt(self, msg)
        try:
            dev_name = msg.arguments[0]
            device = self.factory.devices[dev_name]
        except KeyError:
            return Message.reply('halt', 'fail',
                                 'Unknown device %s' % dev_name)
        else:
            self.factory.unregister_device(dev_name)
            device.send_request('halt').addCallback(got_halt)
            raise AsyncReply()

    def __getattr__(self, attr):
        # TODO: It would be cleaner if a new request method / callback
        # wasn't created for every proxied request.
        # TODO: These proxied methods should appear in the ?help for the proxy
        # but currently don't.

        def request_returned((informs, reply), reqmsg):
            assert informs == []  # for now
            # we *could* in theory just change message name, but let's copy
            # just in case
            self.send_reply(Message.reply(dev_name + "-" + req_name,
                                          *reply.arguments), reqmsg)

        def request_failed(failure, reqmsg):
            self.send_reply(Message.reply(dev_name + '-' + req_name,
                                          "fail", failure.getErrorMessage()),
                            reqmsg)

        def callback(msg):
            if device.state is device.UNSYNCED:
                return Message.reply(dev_name + "-" + req_name, "fail",
                                     "Device not synced")
            d = device.send_request(req_name, *msg.arguments)
            d.addCallbacks(request_returned, request_failed,
                           callbackArgs=(msg,), errbackArgs=(msg,))
            raise AsyncReply()

        if not attr.startswith('request_'):
            return object.__getattribute__(self, attr)
        lst = attr.split('_', 2)
        if len(lst) < 3:
            return object.__getattribute__(self, attr)
        dev_name, req_name = lst[1], lst[2]
        device = self.factory.devices.get(dev_name, None)
        if device is None:
            return object.__getattribute__(self, attr)
        return callback


class ClientDeviceFactory(ClientFactory):
    """ A factory that does uses prebuilt device handler objects
    """
    def __init__(self, addr_mapping, max_reconnects, conn_delay_timeout,
                 proxy):
        self.addr_mapping = addr_mapping  # shared dict with proxy
        self.max_reconnects = max_reconnects
        self.conn_delay_timeout = conn_delay_timeout
        self.proxy = proxy

    def clientConnectionFailed(self, connector, reason):
        addr = connector.host, connector.port
        device = self.addr_mapping[addr]
        if device._conn_counter < self.max_reconnects:
            device._conn_counter += 1
            reactor.callLater(self.conn_delay_timeout,
                              device.schedule_resyncing)
        else:
            self.proxy.devices_scan_failed()

    def buildProtocol(self, addr):
        return self.addr_mapping[(addr.host, addr.port)]


class ProxyKatCP(DeviceServer):
    """ This is a proxy class that will listen on a given host and port
    providing info about underlaying clients if needed
    """
    protocol = ProxyProtocol

    MAX_RECONNECTS = 10
    CONN_DELAY_TIMEOUT = 1

    def __init__(self, *args, **kwds):
        DeviceServer.__init__(self, *args, **kwds)
        self.addr_mapping = {}
        self.client_factory = ClientDeviceFactory(self.addr_mapping,
                                                  self.MAX_RECONNECTS,
                                                  self.CONN_DELAY_TIMEOUT,
                                                  self)
        self.ready_devices = 0
        self.devices = {}
        self.setup_devices()
        self.scan_called = False

    def device_ready(self, device):
        self.ready_devices += 1
        if self.ready_devices == len(self.devices) and not self.scan_called:
            self.scan_called = True  # one shot only
            self.devices_scan_complete()

    def add_device(self, device):
        """ Add a single device to the list of devices that we have
        """
        def really_add_device(addr):
            device.host = addr
            reactor.connectTCP(device.host, device.port, self.client_factory)
            self.devices[device.name] = device
            self.addr_mapping[(device.host, device.port)] = device
            device.add_proxy(self)

        reactor.resolve(device.host).addCallback(really_add_device)

    def add_proxied_sensor(self, device, sensor):
        self.sensors[sensor.name] = sensor

    def devices_scan_complete(self):
        """ A callback called when devices are properly set up and read.
        Override if needed
        """
        pass

    def devices_scan_failed(self):
        """ A callback called when device setup failed to connect after
        repeated tries
        """
        pass

    def setup_devices(self):
        raise NotImplementedError("Override this to provide devices setup")

    def stop(self):
        for device in self.devices.values():
            if device.state != device.UNSYNCED:
                device.stopping = True
                device.stop()
                device.transport.loseConnection()
        DeviceServer.stop(self)

    def unregister_device(self, dev_name):
        device = self.devices[dev_name]
        device.stopping = True
        for name in self.sensors.keys():
            if (name.startswith(dev_name + '.') or
                name.startswith(dev_name + '-')):
                del self.sensors[name]
        del self.devices[dev_name]



# testutils.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 SKA South Africa (http://ska.ac.za/)
# BSD license - see COPYING for details

"""Test utils for katcp package tests.
   """

import client
import logging
import re
import time
import Queue
import threading
import functools
import mock

from .core import Sensor, Message
from .server import (DeviceServer, FailReply, ClientRequestConnection,
                     ClientConnectionTCP)

class ClientConnectionTest(object):
    """
    A version of katcp.server.ClientConnection* suitable for testing
    Records all messages
    """
    def __init__(self):
        self.messages = []

    def inform(self, msg):
        self.messages.append(msg)

    def reply(self, msg, req_msg):
        self.messages.append(msg)

    @property
    def informs(self):
        return [m for m in self.messages if m.mtype == Message.INFORM]

    @property
    def replies(self):
        return [m for m in self.messages if m.mtype == Message.REPLY]

class TestLogHandler(logging.Handler):
    """A logger for KATCP tests."""

    def __init__(self):
        """Create a TestLogHandler."""
        logging.Handler.__init__(self)
        self._records = []

    def emit(self, record):
        """Handle the arrival of a log message."""
        self._records.append(record)

    def clear(self):
        """Clear the list of remembered logs."""
        self._records = []


class DeviceTestSensor(Sensor):
    """Test sensor."""

    def __init__(self, sensor_type, name, description, units, params,
                 timestamp, status, value):
        super(DeviceTestSensor, self).__init__(
            sensor_type, name, description, units, params)
        self.set(timestamp, status, value)


class MessageRecorder(object):
    """
        Parameters
        ----------
        whitelist : list or tuple
            Record only messages matching these names.  If the list is empty,
            all received messages will be saved except any specified in the
            blacklist.
        blacklist : list or tuple
            Ignore messages matching these names.  If any names appear both in
            the whitelist and the blacklist, the whitelist will take
            precedence.
        regex_filter : str
            Record only messages that matches this regexes. Applied to the
            unescaped message string. Messages are pre-filtered by the whitelist
            parameter, so only messages that match both this regexe and have
            names listed in whitelist are recorded.
        informs : boolean
            Whether to record informs.  Default: True.
        replies : boolean
            Whether to record replies.  Default: False.
    """

    def __init__(self, msg_types, whitelist, regex_filter, blacklist):
        self.msgs = []
        self.msg_types = msg_types
        self.whitelist = whitelist
        self.blacklist = blacklist
        self.regex_filter = re.compile(regex_filter) if regex_filter else None
        self.msg_received = threading.Event()

    def get_msgs(self, min_number=0, timeout=1):
        """Return the messages recorded so far. and delete them

        Parameters
        ==========
        number -- Minumum number of messages to return, else wait
        timeout: int seconds or None -- Don't wait longer than this
        """
        self.wait_number(min_number, timeout)
        msg_count = len(self.msgs)
        msgs_copy = self.msgs[:msg_count]
        del self.msgs[:msg_count]
        return msgs_copy

    __call__ = get_msgs           # For backwards compatibility

    def wait_number(self, number, timeout=1):
        """
        Wait until at least certain number of messages have been recorded

        Parameters
        ==========
        number -- Number of messages to wait for
        timeout: int seconds or None -- Don't wait longer than this
        """
        start_time = time.time()
        while True:
            if len(self.msgs) >= number:
                return True
            elapsed_time = time.time() - start_time
            if elapsed_time >= timeout:
                return None
            self.msg_received.wait(timeout=timeout-elapsed_time)

    def append_msg(self, msg):
        """Append a message if it matches the criteria."""
        if msg.mtype not in self.msg_types:
            return
        try:
            if self._record_predicate(msg):
                self.msgs.append(msg)
                self.msg_received.set()
        finally:
            self.msg_received.clear()

    def _record_predicate(self, msg):
        if self.whitelist and msg.name not in self.whitelist:
            return False
        if self.regex_filter:
            if self.regex_filter.match(str(msg)):
                return True
            else:
                return False
        if msg.name in self.blacklist:
            return False
        return True


class BlockingTestClient(client.BlockingClient):
    """Test blocking client."""

    def __init__(self, test, *args, **kwargs):
        """Takes a TestCase class as an additional parameter."""
        self.test = test
        self._message_recorders = {}
        super(BlockingTestClient, self).__init__(*args, **kwargs)

    def raw_send(self, chunk):
        """Send a raw chunk of data to the server."""
        self._sock.send(chunk)

    def _sensor_lag(self):
        """The expected lag before device changes are applied."""
        return getattr(self.test, "sensor_lag", 0)

    def handle_inform(self, msg):
        """Pass unhandled informs to message recorders."""
        for append_msg in self._message_recorders.itervalues():
            append_msg(msg)
        return super(BlockingTestClient, self).handle_inform(msg)

    def handle_reply(self, msg):
        """Pass unhandled replies to message recorders."""
        for append_msg in self._message_recorders.itervalues():
            append_msg(msg)
        return super(BlockingTestClient, self).handle_reply(msg)

    def message_recorder(self, whitelist=(), blacklist=(), regex_filter=None,
                         informs=True, replies=False):
        """Helper method for attaching a hook to selected received messages.

        Parameters
        ----------
        whitelist : list or tuple
            Record only messages matching these names.  If the list is empty,
            all received messages will be saved except any specified in the
            blacklist.
        blacklist : list or tuple
            Ignore messages matching these names.  If any names appear both in
            the whitelist and the blacklist, the whitelist will take
            precedence.
        regex_filter : str
            Record only messages that matches this regexes. Applied to the
            unescaped message string. Messages are pre-filtered by the whitelist
            parameter, so only messages that match both this regexe and have
            names listed in whitelist are recorded.
        informs : boolean
            Whether to record informs.  Default: True.
        replies : boolean
            Whether to record replies.  Default: False.

        Return
        ------
        get_msgs : object
            Callable Instance of MessageRecorder. Returns a list of messages
            that have matched so far.  Each call returns the list of message
            since the previous call.
        """
        msg_types = set()
        if informs:
            msg_types.add(Message.INFORM)
        if replies:
            msg_types.add(Message.REPLY)
        whitelist = set(whitelist)
        blacklist = set(blacklist)

        mr = MessageRecorder(msg_types, whitelist, regex_filter, blacklist)
        self._message_recorders[id(mr)] = mr.append_msg

        return mr

    @staticmethod
    def expected_sensor_value_tuple(sensorname, value, sensortype=str,
                                    places=7):
        """Helper method for completing optional values in expected sensor
        value tuples.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        value : obj
            The expected value of the sensor.  Type must match sensortype.
        sensortype : type, optional
            The type to use to convert the sensor value. Default: str.
        places : int, optional
            The number of places to use in a float comparison.  Has no effect
            if sensortype is not float. Default: 7.
        """

        return (sensorname, value, sensortype, places)

    # SENSOR VALUES

    def get_sensor(self, sensorname, sensortype=str):
        """Retrieve the value, status and timestamp of a sensor.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        sensortype : type, optional
            The type to use to convert the sensor value. Default: str.
        """

        reply, informs = self.blocking_request(Message.request("sensor-value",
                                                               sensorname))
        self.test.assertTrue(reply.reply_ok(),
            "Could not retrieve sensor '%s': %s"
            % (sensorname, (reply.arguments[1] if len(reply.arguments) >= 2
                            else "")))

        self.test.assertEqual(len(informs), 1,
            "Expected one sensor value inform for sensor '%s', but"
            " received %d." % (sensorname, len(informs)))

        inform = informs[0]

        self.test.assertEqual(len(inform.arguments), 5,
            "Expected sensor value inform for sensor '%s' to have five"
            " arguments, but received %d." %
            (sensorname, len(inform.arguments)))

        timestamp, _num, _name, status, value = inform.arguments

        try:
            if sensortype == bool:
                typestr = "%r and then %r" % (int, bool)
                value = bool(int(value))
            else:
                typestr = "%r" % sensortype
                value = sensortype(value)
        except ValueError, e:
            self.test.fail("Could not convert value %r of sensor '%s' to type"
                           " %s: %s" % (value, sensorname, typestr, e))

        self.test.assertTrue(status in Sensor.STATUSES.values(),
                "Got invalid status value %r for sensor '%s'." %
                (status, sensorname))

        try:
            timestamp = float(timestamp)
        except ValueError, e:
            self.test.fail("Could not convert timestamp %r of sensor '%s' to"
                           " type %r: %s" % (timestamp, sensorname, float, e))

        return value, status, timestamp

    def get_sensor_value(self, sensorname, sensortype=str):
        """Retrieve the value of a sensor.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        sensortype : type, optional
            The type to use to convert the sensor value. Default: str.
        """

        value, _, _ = self.get_sensor(sensorname, sensortype)
        return value

    def get_sensor_status(self, sensorname):
        """Retrieve the status of a sensor.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        """

        _, status, _ = self.get_sensor(sensorname)
        return status

    def get_sensor_timestamp(self, sensorname):
        """Retrieve the timestamp of a sensor.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        """

        _, _, timestamp = self.get_sensor(sensorname)
        return timestamp

    def assert_sensor_equals(self, sensorname, expected, sensortype=str,
                             msg=None, places=7, status=None):
        """Assert that a sensor's value is equal to the given value.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        expected : obj
            The expected value of the sensor.  Type must match sensortype.
        sensortype : type, optional
            The type to use to convert the sensor value. Default: str.
        msg : str, optional
            A custom message to print if the assertion fails.  If the string
            contains %r, it will be replaced with the sensor's value. A default
            message is defined in this method.
        places : int, optional
            The number of places to use in a float comparison.  Has no effect
            if sensortype is not float. Default: 7.
        status : str, optional
            The expected status of the sensor.
        """

        if msg is None:
            places_msg = " (within %d decimal places)" % \
                         places if sensortype == float else ""
            msg = "Value of sensor '%s' is %%r. Expected %r%s." % \
                  (sensorname, expected, places_msg)

        got, got_status, _timestamp = self.get_sensor(sensorname, sensortype)
        if '%r' in msg:
            msg = msg % got

        if sensortype == float:
            self.test.assertAlmostEqual(got, expected, places, msg)
        else:
            self.test.assertEqual(got, expected, msg)

        if status is not None:
            self.test.assertEqual(got_status, status, "Status of sensor '%s'"
                                  " is %r. Expected %r." % (sensorname,
                                                            got_status,
                                                            status))

    def assert_sensor_status_equals(self, sensorname, expected_status,
                                    msg=None):
        """Assert that a sensor's status is equal to the given status.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        expected_status : str
            The expected status of the sensor.
        msg : str, optional
            A custom message to print if the assertion fails.  If the string
            contains %r, it will be replaced with the sensor's value. A default
            message is defined in this method.
        """

        if msg is None:
            msg = "Status of sensor '%s' is %%r. Expected %r." % \
                  (sensorname, expected_status)

        got_status = self.get_sensor_status(sensorname)

        if '%r' in msg:
            msg = msg % got_status

        self.test.assertEqual(got_status, expected_status, msg)

    def assert_sensor_not_equal(self, sensorname, expected, sensortype=str,
                                msg=None, places=7):
        """Assert that a sensor's value is not equal to the given value.

        Parameters
        ----------
        sensorname : str
            The name of the sensor.
        expected : obj
            The expected value of the sensor.  Type must match sensortype.
        sensortype : type, optional
            The type to use to convert the sensor value. Default: str.
        msg : str, optional
            A custom message to print if the assertion fails.  If the string
            contains %r, it will be replaced with the sensor's value. A default
            message is defined in this method.
        places : int, optional
            The number of places to use in a float comparison.  Has no effect
            if sensortype is not float. Default: 7.
        """

        if msg is None:
            places_msg = " (within %d decimal places)" % \
                         places if sensortype == float else ""
            msg = "Value of sensor '%s' is %%r. Expected a different" \
                  " value%s." % (sensorname, places_msg)

        got = self.get_sensor_value(sensorname, sensortype)
        if '%r' in msg:
            msg = msg % got

        if sensortype == float:
            self.test.assertNotAlmostEqual(got, expected, places, msg)
        else:
            self.test.assertNotEqual(got, expected, msg)

    def assert_sensors_equal(self, sensor_tuples, msg=None):
        """Assert that the values of several sensors are equal to the given
        values.

        Parameters
        ----------
        sensor_tuples : list of tuples
            A list of tuples specifying the sensor values to be checked.  See
            :meth:`expected_sensor_value_tuple`.
        msg : str, optional
            A custom message to print if the assertion fails.  If the string
            contains %r, it will be replaced with the sensor's value. A default
            message is defined in this method.
        """

        sensor_tuples = [self.expected_sensor_value_tuple(*t)
                         for t in sensor_tuples]
        for sensorname, expected, sensortype, places in sensor_tuples:
            self.assert_sensor_equals(sensorname, expected, sensortype,
                                      msg=msg, places=places)

    def assert_sensors_not_equal(self, sensor_tuples, msg=None):
        """Assert that the values of several sensors are not equal to the given
        values.

        Parameters
        ----------
        sensor_tuples : list of tuples
            A list of tuples specifying the sensor values to be checked.  See
            :meth:`expected_sensor_value_tuple`.
        msg : str, optional
            A custom message to print if the assertion fails.  If the string
            contains %r, it will be replaced with the sensor's value. A default
            message is defined in this method.
        """

        sensor_tuples = [self.expected_sensor_value_tuple(*t)
                         for t in sensor_tuples]
        for sensorname, expected, sensortype, places in sensor_tuples:
            self.assert_sensor_not_equal(sensorname, expected, sensortype,
                                         msg=msg, places=places)

    def wait_until_sensor_equals(self, timeout, sensorname, value,
                                 sensortype=str, places=7, pollfreq=0.02):
        """Wait until a sensor's value is equal to the given value, or time
        out.

        Parameters
        ----------
        timeout : float
            How long to wait before timing out, in seconds.
        sensorname : str
            The name of the sensor.
        value : obj
            The expected value of the sensor.  Type must match sensortype.
        sensortype : type, optional
            The type to use to convert the sensor value. Default: str.
        places : int, optional
            The number of places to use in a float comparison.  Has no effect
            if sensortype is not float. Default: 7.
        pollfreq : float, optional
            How frequently to poll for the sensor value. Default: 0.1.
        """

        # TODO Should be changed to use some varient of SensorTransitionWaiter

        stoptime = time.time() + timeout
        success = False

        if sensortype == float:
            cmpfun = lambda got, exp: abs(got - exp) < 10 ** -places
        else:
            cmpfun = lambda got, exp: got == exp

        lastval = None
        while time.time() < stoptime:
            lastval = self.get_sensor_value(sensorname, sensortype)
            if cmpfun(lastval, value):
                success = True
                break
            time.sleep(pollfreq)

        if not success:
            self.test.fail("Timed out while waiting %ss for %s sensor to"
                           " become %s. Last value was %s." %
                           (timeout, sensorname, value, lastval))

    # REQUEST PARAMETERS

    def test_sensor_list(self, expected_sensors, ignore_descriptions=False):
        """Test that the list of sensors on the device equals the provided
        list.

        Parameters
        ----------
        expected_sensors : list of tuples
            The list of expected sensors.  Each tuple contains the arguments
            returned by each sensor-list inform, as unescaped strings.
        ignore_descriptions : boolean, optional
            If this is true, sensor descriptions will be ignored in the
            comparison. Default: False.
        """

        def sensortuple(name, description, units, stype, *params):
            # ensure float params reduced to the same format
            if stype == "float":
                params = ["%g" % float(p) for p in params]
            return (name, description, units, stype) + tuple(params)

        reply, informs = self.blocking_request(Message.request("sensor-list"))

        self.test.assertTrue(reply.reply_ok(),
            "Could not retrieve sensor list: %s"
            % (reply.arguments[1] if len(reply.arguments) >= 2 else ""))

        expected_sensors = [sensortuple(*t) for t in expected_sensors]
        got_sensors = [sensortuple(*m.arguments) for m in informs]

        #print ",\n".join([str(t) for t in got_sensors])

        if ignore_descriptions:
            expected_sensors = [s[:1] + s[2:] for s in expected_sensors]
            got_sensors = [s[:1] + s[2:] for s in got_sensors]

        expected_set = set(expected_sensors)
        got_set = set(got_sensors)

        self.test.assertEqual(got_set, expected_set,
            "Sensor list differs from expected list.\nThese sensors are"
            " missing:\n%s\nFound these unexpected sensors:\n%s"
            % ("\n".join(sorted([str(t) for t in expected_set - got_set])),
               "\n".join(sorted([str(t) for t in got_set - expected_set]))))

    def test_help(self, expected_requests, full_descriptions=False,
                  exclude_defaults=True):
        """Test that the list of requests on the device equals the provided
        list.

        Parameters
        ----------
        expected_requests : list of tuples or strings
            The list of expected requests.  May be a list of request names as
            strings.  May also be a list of tuples, each of which contains the
            request name and either the first line or the full text of the
            request description, as unescaped strings.  If full_descriptions is
            true, full description text must be provided.
        full_descriptions : boolean, optional
            If this is true, the full text of descriptions is compared,
            otherwise only text up to the first newline is compared.  Has no
            effect if expected_requests is a list of strings.  Default: False.
        exclude_defaults : boolean, optional
            If this is true, exclude requests on the device which match
            requests found on :class:`katcp.DeviceServer` by name.
            Default: True.
        """

        descriptions = True
        if not expected_requests or isinstance(expected_requests[0], str):
            descriptions = False

        reply, informs = self.blocking_request(Message.request("help"))
        self.test.assertTrue(reply.reply_ok(),
            "Could not retrieve help list: %s"
            % (reply.arguments[1] if len(reply.arguments) >= 2 else ""))

        got_requests = [tuple(m.arguments) for m in informs]

        if descriptions:
            if not full_descriptions:
                got_requests = [(name, desc.split('\n')[0])
                                for (name, desc) in got_requests]
                expected_requests = [(name, desc.split('\n')[0])
                                     for (name, desc) in expected_requests]
        else:
            got_requests = [name for (name, desc) in got_requests]

        got_set = set(got_requests)
        expected_set = set(expected_requests)

        if exclude_defaults:
            default_request_names = set(DeviceServer._request_handlers.keys())

            if not descriptions:
                got_set = got_set - default_request_names
            else:
                got_set = set([(name, desc) for (name, desc) in got_set
                               if name not in default_request_names])

        self.test.assertEqual(got_set, expected_set,
            "Help list differs from expected list.\nThese requests are"
            " missing:\n%s\nFound these unexpected requests:\n%s"
            % ("\n".join(sorted([str(t) for t in expected_set - got_set])),
               "\n".join(sorted([str(t) for t in got_set - expected_set]))))

    def assert_request_succeeds(self, requestname, *params, **kwargs):
        """Assert that the given request completes successfully when called
        with the given parameters, and optionally check the arguments.

        Parameters
        ----------
        requestname : str
            The name of the request.
        params : list of objects
            The parameters with which to call the request.
        args_echo : boolean, optional
            Keyword parameter.  Assert that the reply arguments after 'ok'
            equal the request parameters.  Takes precedence over args_equal,
            args_in and args_length. Default False.
        args_equal : list of strings, optional
            Keyword parameter.  Assert that the reply arguments after 'ok'
            equal this list.  Ignored if args_echo is present; takes precedence
            over args_echo, args_in and args_length.
        args_in : list of lists, optional
            Keyword parameter.  Assert that the reply arguments after 'ok'
            equal one of these tuples.  Ignored if args_equal or args_echo is
            present; takes precedence over args_length.
        args_length : int, optional
            Keyword parameter.  Assert that the length of the reply arguments
            after 'ok' matches this number. Ignored if args_equal, args_echo or
            args_in is present.
        informs_count : int, optional
            Keyword parameter.  Assert that the number of informs received
            matches this number.
        """

        reply, informs = self.blocking_request(Message.request(requestname,
                                                               *params))

        self.test.assertEqual(reply.name, requestname, "Reply to request '%s'"
                              " has name '%s'." % (requestname, reply.name))

        self.test.assertTrue(reply.reply_ok(),
            "Expected request '%s' called with parameters %r to succeed, but"
            " it failed %s." %
            (requestname, params, ("with error '%s'" % reply.arguments[1]
                                   if len(reply.arguments) >= 2
                                   else "(with no error message)")))

        args = reply.arguments[1:]

        args_echo = kwargs.get("args_echo", False)
        args_equal = kwargs.get("args_equal")
        args_in = kwargs.get("args_in")
        args_length = kwargs.get("args_length")
        informs_count = kwargs.get("informs_count")

        if args_echo:
            args_equal = [str(p) for p in params]

        if args_equal is not None:
            self.test.assertEqual(args, args_equal,
                "Expected reply to request '%s' called with parameters %r to"
                " have arguments %s, but received %s."
                % (requestname, params, args_equal, args))
        elif args_in is not None:
            self.test.assertTrue(args in args_in,
                "Expected reply to request '%s' called with parameters %r to"
                " have arguments in %s, but received %s."
                % (requestname, params, args_in, args))
        elif args_length is not None:
            self.test.assertEqual(len(args), args_length,
                "Expected reply to request '%s' called with parameters %r to"
                " have %d arguments, but received %d: %s."
                % (requestname, params, args_length, len(args), args))

        if informs_count is not None:
            self.test.assertEqual(len(informs), informs_count,
                "Expected %d informs in reply to request '%s' called with"
                " parameters %r, but received %d."
                % (informs_count, requestname, params, len(informs)))

        return args

    def assert_request_fails(self, requestname, *params, **kwargs):
        """Assert that the given request fails when called with the given
        parameters.

        Parameters
        ----------
        requestname : str
            The name of the request.
        params : list of objects
            The parameters with which to call the request.
        status_equals : string, optional
            Keyword parameter.  Assert that the reply status equals this
            string.
        error_equals : string, optional
            Keyword parameter.  Assert that the error message equals this
            string.
        """

        reply, informs = self.blocking_request(Message.request(requestname,
                                                               *params))

        self.test.assertEqual(reply.name, requestname,
                              "Reply to request '%s' has name '%s'." %
                              (requestname, reply.name))

        self.test.assertFalse(reply.reply_ok(), "Expected request '%s' called"
                              " with parameters %r to fail, but it was"
                              " successful." % (requestname, params))

        status = reply.arguments[0]
        error = reply.arguments[1] if len(reply.arguments) > 1 else None

        status_equals = kwargs.get("status_equals")
        error_equals = kwargs.get("error_equals")

        if status_equals is not None:
            self.test.assertTrue(status == status_equals,
                "Expected request '%s' called with parameters %r to return"
                " status %s, but the status was %r."
                % (requestname, params, status_equals, status))

        if error_equals is not None:
            self.test.assertTrue(error is not None and error == error_equals,
                "Expected request '%s' called with parameters %r to fail with"
                " error %s, but the error was %r."
                % (requestname, params, error_equals, error))

        return status, error

    def test_setter_request(self, requestname, sensorname, sensortype=str,
                            good=(), bad=(), places=7, args_echo=False):
        """Test a request which simply sets the value of a sensor.

        Parameters
        ----------
        requestname : str
            The name of the request.
        sensorname : str
            The name of the sensor.
        sensortype : type, optional
            The type to use to convert the sensor value. Default: str.
        good : list of objects
            A list of values to which this request can successfully set the
            sensor.  The object type should match sensortype.
        bad : list of objects
            A list of values to which this request cannot successfully set the
            sensor.  The object type should match sensortype.
        places : int, optional
            The number of places to use in a float comparison.  Has no effect
            if sensortype is not float. Default: 7.
        args_echo : boolean, optional
            Check that the value is echoed as an argument by the reply to the
            request.  Default: False.
        """

        for value in good:
            self.assert_request_succeeds(requestname, value,
                                         args_echo=args_echo)
            time.sleep(self._sensor_lag())

            self.assert_sensor_equals(sensorname, value, sensortype,
                "After request '%s' was called with parameter %r, value of"
                " sensor '%s' is %%r. Expected %r%s."
                % (requestname, value, sensorname, value,
                   (" (within %d decimal places)" % places
                    if sensortype == float else "")), places)

        for value in bad:
            self.assert_request_fails(requestname, value)

    def test_multi_setter_request(self, requestname, good=(), bad=()):
        """Test a request which causes several sensor values to change.

        Parameters
        ----------
        requestname : str
            The name of the request.
        good : list of tuples
            Each tuple contains a tuple of successful parameters, a tuple of
            expected sensor values (see :meth:`expected_sensor_value_tuple`),
            and optionally a dict of options. Permitted options are: "statuses"
            and a list of status tuples to check, or "delay" and a float in
            seconds specifying an additional delay before the sensors are
            expected to change.
        bad : list of tuples
            Each tuple is set of parameters which should cause the request to
            fail.
        """

        def testtuple(params, expected_values, options={}):
            return (params, expected_values, options)

        good = [testtuple(*t) for t in good]

        for params, expected_values, options in good:
            delay = options.get("delay", 0)
            expected_statuses = options.get("statuses", ())

            self.assert_request_succeeds(requestname, *params)
            time.sleep(self._sensor_lag() + delay)

            expected_values = [self.expected_sensor_value_tuple(*t)
                               for t in expected_values]

            for sensorname, value, sensortype, places in expected_values:
                self.assert_sensor_equals(sensorname, value, sensortype,
                    "After request '%s' was called with parameters %r, value"
                    " of sensor '%s' is %%r. Expected %r%s."
                    % (requestname, params, sensorname, value,
                       (" (within %d decimal places)" % places
                        if sensortype == float else "")), places)

            for sensorname, status in expected_statuses:
                self.assert_sensor_status_equals(sensorname, status,
                    "After request '%s' was called with parameters %r, status"
                    " of sensor '%s' is %%r. Expected %r."
                    % (requestname, params, sensorname, status))

        for params in bad:
            self.assert_request_fails(requestname, *params)


class DeviceTestServer(DeviceServer):
    """Test server."""

    def __init__(self, *args, **kwargs):
        super(DeviceTestServer, self).__init__(*args, **kwargs)
        self.__msgs = []
        self.restart_queue = Queue.Queue()
        self.set_restart_queue(self.restart_queue)
        self._cancel_slow_command = threading.Event()
        self.slow_waiting = False

    def setup_sensors(self):
        self.restarted = False
        self.add_sensor(DeviceTestSensor(
            Sensor.INTEGER, "an.int", "An Integer.", "count",
            [-5, 5],
            timestamp=12345, status=Sensor.NOMINAL, value=3))

    def request_new_command(self, req, msg):
        """A new command."""
        return Message.reply(msg.name, "ok", "param1", "param2")

    def request_raise_exception(self, req, msg):
        """A handler which raises an exception."""
        raise Exception("An exception occurred!")

    def request_raise_fail(self, req, msg):
        """A handler which raises a FailReply."""
        raise FailReply("There was a problem with your request.")

    def request_slow_command(self, req, msg):
        """A slow command, waits for msg.arguments[0] seconds"""
        self.slow_waiting = True
        self._cancel_slow_command.wait(float(msg.arguments[0]))
        self.slow_waiting = False
        return req.make_reply("ok")

    def request_cancel_slow_command(self, req, msg):
        """Cancel slow command request, resulting in it replying immedietely"""
        self._cancel_slow_command.set()
        return req.make_reply('ok')

    def handle_message(self, req, msg):
        self.__msgs.append(msg)
        super(DeviceTestServer, self).handle_message(req, msg)

    def messages(self):
        return self.__msgs

    def stop(self, *args, **kwargs):
        # Make sure a slow command with long timeout does not hold us
        # up.
        self._cancel_slow_command.set()
        super(DeviceTestServer, self).stop(*args, **kwargs)


class TestUtilMixin(object):
    """Mixin class implementing test helper methods for making
       assertions about lists of KATCP messages.
       """

    def _assert_msgs_length(self, actual_msgs, expected_number):
        """Assert that the number of messages is as expected."""
        num_msgs = len(actual_msgs)
        if num_msgs < expected_number:
            self.assertEqual(num_msgs, expected_number,
                             "Too few messages received.")
        elif num_msgs > expected_number:
            self.assertEqual(num_msgs, expected_number,
                             "Too many messages received.")

    def _assert_msgs_equal(self, actual_msgs, expected_msgs):
        """Assert that the actual and expected messages are equal.

           actual_msgs: list of message objects received
           expected_msgs: expected message strings
           """
        for msg, msg_str in zip(actual_msgs, expected_msgs):
            self.assertEqual(str(msg), msg_str)
        self._assert_msgs_length(actual_msgs, len(expected_msgs))

    def _assert_msgs_match(self, actual_msgs, expected):
        """Assert that the actual messages match the expected regular
           expression patterns.

           actual_msgs: list of message objects received
           expected: expected patterns
           """
        for msg, pattern in zip(actual_msgs, expected):
            self.assertTrue(re.match(pattern, str(msg)),
                            "Message did match pattern %r: %s" %
                            (pattern, msg))
        self._assert_msgs_length(actual_msgs, len(expected))

    def _assert_msgs_like(self, actual_msgs, expected):
        """Assert that the actual messages start and end with
           the expected strings.

           actual_msgs: list of message objects received
           expected_msgs: tuples of (expected_prefix, expected_suffix)
           """
        for msg, (prefix, suffix) in zip(actual_msgs, expected):
            str_msg = str(msg)

            if prefix and not str_msg.startswith(prefix):
                self.assertEqual(str_msg, prefix,
                    msg="Message '%s' does not start with '%s'."
                    % (str_msg, prefix))

            if suffix and not str_msg.endswith(suffix):
                self.assertEqual(str_msg, suffix,
                    msg="Message '%s' does not end with '%s'."
                    % (str_msg, suffix))
        self._assert_msgs_length(actual_msgs, len(expected))

def counting_callback(event=None, number_of_calls=1):
    """Decorate a callback to set an event once it has been called a certain number of times

    Parameters
    ==========
    event: threading.Event() -- will be set when enough calls have been made.
        If None, a new event will be created
    number_of_calls: int > 0 -- Number of calls before event.set() is called

    Decorated Properties
    ====================

    done -- The event object
    get_no_calls() -- Returns current number of calls
    wait(timeout=None) -- Wait for number_of_calls to be made
    assert_wait(timeout=None) -- Wait for number_of_calls and raises
        AssertionError if they are not made before the timeout.
    reset() -- Set call count back to zero
    """
    if event is None:
        event = threading.Event()

    def decorator(original_callback):
        assert number_of_calls > 0
        calls = [0]

        @functools.wraps(original_callback)
        def wrapped_callback(*args, **kwargs):
            retval = original_callback(*args, **kwargs)
            calls[0] += 1
            if calls[0] >= number_of_calls:
                event.set()
            return retval

        wrapped_callback.get_no_calls = lambda : calls[0]
        wrapped_callback.done = event
        wrapped_callback.wait = event.wait

        def assert_wait(timeout=None):
            done = event.wait(timeout)
            assert event.isSet()
            return done
        wrapped_callback.assert_wait = assert_wait

        def reset():
            calls[0] = 0
            event.clear()
        wrapped_callback.reset = reset

        return wrapped_callback
    return decorator


def suppress_queue_repeats(queue, initial_value, read_time=None):
    """Generator that reads a Queue.Queue() and suppresses runs of repeated values

    The queue is consumed, and a value yielded whenever it differs from the
    previous value. If read_time is specified, stops iteration if the
    queue is empty after reading the queue for read_time seconds. read_time=None
    continues reading forever.
    """
    start_time = time.time()
    cur_value = initial_value
    next_wait = read_time

    while True:
        if next_wait:
            next_wait = read_time - (time.time() - start_time)
            next_wait = max(0, next_wait)
        try:
            next_value = queue.get(timeout=next_wait)
        except Queue.Empty:
            break
        if next_value != cur_value:
            yield next_value
            cur_value = next_value

class SensorTransitionWaiter(object):
    """
    Wait for a given set of sensor transitions

    Can be used to test sensor transitions indepedent of timing. If the
    SensorTransitionWaiter object is instantiated before the test is triggered
    the transitions are guaranteed (I hope) to be detected.
    """
    def __init__(self, sensor, value_sequence=None):
        """
        Parameters
        ----------
        sensor : KATSensor object to watch
        value_sequence : list of sensor values or callable conditions or None

            Will check that this sequence of values occurs on the watched
            sensor. The first value is the starting value -- an error will be
            raised if that is not the initial value.  If an object in the list
            is a callable, it will be called on the actual sensor value. If it
            returns True, the value will be considered as matched. If
            value_sequence is None, no checking will be done, and the .wait()
            method cannot be called. However, the received values can be
            retrieved using get_received_values().

        If an exact value sequence is passed, the sensor is watched
        until its value changes. If the sensor does not take on the
        exact sequence of values, the transition is not matched. If
        expected values are repeated (e.g. [1, 2, 2, 3]) the repeated
        values (i.e. the second 2 in the example) are ignored.


        The following algorithm is used to test a For a sequence for validity
        when callable tests are used:

        * The initial value needs to satisfy at least the first test
        * While the current and not next test passes, continue reading
        * If neither current nor next passes, signal failure
        * If next passes, move test sequence along

        This can be used to check a sequence of floating point values:

        waiter = WaitForSensorTransition(sensor, value_sequence=(
            lambda x: x < 0.7,
            lambda x: x >= 0.7,
            lambda x: x >= 1,
            lambda x: x < 1,
            lambda x: x < 0.3)
        waiter.wait(timeout=1)

        can be used to check that a sensor starts at a value smaller than 0.7,
        then grows to a value larger than 1 and then shrinks to a value smaller
        than 0.3. Note that the x >= 0.7 test is required in case the sensor
        value attains exactly 0.7.

        For tests like this it is important to specify a timeout.
        """

        self.sensor = sensor
        self.desired_value_sequence = value_sequence
        self._torn_down = False
        self._done = False
        self._value_queue = Queue.Queue()
        self.timed_out = False
        current_value = self._get_current_sensor_value()
        if value_sequence:
            if not self._test_value(current_value, value_sequence[0]):
                raise ValueError('Sensor value does not satisfy initial condition')
        self.received_values = [current_value]
        self._configure_sensor()

    def _get_current_sensor_value(self):
        return self.sensor.value()

    def _configure_sensor(self):
        # Do neccesary sensor configuration. Assumes sensor is stored as
        # self.sensor

        # Attach our observer to the katcp sensor
        class observer(object): pass
        observer.update = self._sensor_callback
        self._observer = observer
        self.sensor.attach(self._observer)

    def _teardown_sensor(self):
        # Perform teardown cleanup actions on self.sensor
        self.sensor.detach(self._observer)

    def _sensor_callback(self, sensor):
        assert(sensor is self.sensor)
        self._value_queue.put(sensor.value())

    def _test_value(self, value, value_test):
        """
        Test value against value_test

        value_test can be a callable or a simple value. If it is a simple
        value it is compared for equality, otherwise value_test(value)
        is called and the result returned
        """
        if callable(value_test):
            return value_test(value)
        else:
            return value == value_test

    def wait(self, timeout=5):
        """
        Wait until the specified transition occurs

        Parameters
        ----------
        timeout : float seconds or None
            Time to wait for the transition. Wait forever if None

        Returns
        -------

        Returns True if the sequence is matched within the
        timeout. Returns False if the sequence does not match or if
        the timeout expired. The actual received values is available
        as the `received_values` member of the object. Sets the member
        `timed_out` to True if a timeout occured.
        """
        if self._done:
            raise RuntimeError('Transition already triggered. Instantiate a new '
                               'SensorTransitionWaiter object')
        if self._torn_down:
            raise RuntimeError('This object has been torn down. Instantiate a new '
                               'SensorTransitionWaiter object')
        nonrepeat_sensor_values = suppress_queue_repeats(
            self._value_queue, self.received_values[-1], timeout)

        try:
            # While current and not next test passes, continue reading
            # If neither current nor next passes, signal failure
            # If next passes, move test sequence along
            for current_test, next_test in zip(
                    self.desired_value_sequence[:-1],
                    self.desired_value_sequence[1:]):
                if self._test_value(self.received_values[-1], next_test):
                    continue # Next test already satisfied by current value
                while True:
                    # Read values from the queue until either the timeout
                    # expires or a value different from the last is found
                    next_value = nonrepeat_sensor_values.next()
                    self.received_values.append(next_value)
                    current_pass = self._test_value(next_value, current_test)
                    next_pass = self._test_value(next_value, next_test)
                    if next_pass:
                        break         # Matches, move on to the next test
                    if not current_pass:
                        # Matches neither the current test nor the
                        # next. Indicates an invalid sequence.
                        return False
            # We have passed all test conditions, hence the desired
            # transition has occured
            return True
        except StopIteration:
            self.timed_out = True
            return False
        finally:
            self._done = True
            self.teardown()

    def get_received_values(self, stop=True, reset=True):
        """Return list of values received to date.

        Parameters
        ----------

        stop -- Stop listening, tear down and unsubscribe from sensor
        reset -- Reset the recieved values to an empty list

        Note, once this method has been called, wait() can no longer work
        """
        try:
            while True:
                self.received_values.append(self._value_queue.get_nowait())
        except Queue.Empty:
            pass
        received_values = self.received_values
        if reset:
            self.received_values = []
        if stop:
            self.teardown()
        return received_values

    def teardown(self):
        """Clean up, restoring sensor strategy and deregistering the sensor callback"""
        if self._torn_down:
            return
        self._teardown_sensor()
        self._torn_down = True

def wait_sensor(sensor, value, timeout=5):
    """Wait for a katcp sensor to attain a certain value

    Temporarily attaches to the sensor to get sensor updates. It is assumed that
    the sensor is getting updated by another thread.

    """
    tests = (lambda x: True, value)
    waiter = SensorTransitionWaiter(sensor, tests)
    return waiter.wait(timeout=timeout)


def start_thread_with_cleanup(
        test_instance, thread_object, timeout=1, start_timeout=None):
    """Start thread_object and add cleanup functions to test_instance

    thread_object.start() is called to start the thread.
    thread_object.join(timeout=timeout) and thread_object.stop() is added to the
    test instance cleanup
    """
    if start_timeout is not None:
        thread_object.start(timeout=timeout)
    else:
        thread_object.start()
    test_instance.addCleanup(thread_object.join, timeout=timeout)
    test_instance.addCleanup(thread_object.stop)

class WaitingMock(mock.Mock):
    def __init__(self, *args, **kwargs):
        super(WaitingMock, self).__init__(*args, **kwargs)
        self._wait_call_count_event = threading.Event()
        self._num_calls_waiting_for = None

    def _mock_call(self, *args, **kwargs):
        retval = super(WaitingMock, self)._mock_call(*args, **kwargs)
        waiting_for = self._num_calls_waiting_for
        if waiting_for and self.call_count >= waiting_for:
            self._wait_call_count_event.set()
        return retval

    def assert_wait_call_count(self, count, timeout=1.):
        """
        Wait for mock to be called at least 'count' times within 'timeout' seconds

        Raises AssertionError if the call count is not reached.
        """
        self._num_calls_waiting_for = count
        self._wait_call_count_event.clear()
        if self.call_count >= count:
            return True
        self._wait_call_count_event.wait(timeout=timeout)
        assert(self._wait_call_count_event.isSet())

def mock_req(req_name, *args, **kwargs):
    """
    Create a mock ClientRequestConnection object

    Parameters
    ----------

    req_name : str
        Name of the request
    *args : obj
        arguments for the request, used to construct a request Message object

    Optional Keyword Arguments
    --------------------------

    server : obj
       Used as the server instance when constructing a client_connection.
       Cannot be specified together with client_conn
    client_conn : obj
       Used as the client_connection object when constructing mock
       ClientReqeuestConnection object. Cannot be specified together with sock
    sock : obj
       Used as the socket object when constructing a server and client_connection

    If server but not sock is specified, a mock sock and real a
    ClientConnectionTCP instances are used. If client_conn is not specified, a
    WaitingMock instance is used instead.

    Returns
    -------

    req : WaitingMock instance
        The `client_connection` and `msg` attributes are set. The
        :meth:`make_reply()` method returns the appropriate request Message
        object as a side effect.
    """

    server = kwargs.get('server')
    client_conn = kwargs.get('client_conn')
    sock = kwargs.get('sock')
    if server and client_conn:
        raise TypeError('Specify either server or client_conn, not both')
    if client_conn and sock:
        raise TypeError('Specify either client_conn or sock, not both')
    if not sock:
        sock = WaitingMock()
    if server:
        client_conn = ClientConnectionTCP(server, sock)
    if not client_conn:
        client_conn = WaitingMock()

    req = WaitingMock()
    req.client_connection = client_conn
    req.msg = Message.request(req_name, *args)
    req.make_reply.side_effect = lambda *args: Message.reply_to_request(
        req.msg, *args)
    return req
